/*****************************************************************************
 * Lemonade JS
 * https://lemonadejs.com
 *
 * ----
 *
 * Cassandra
 *
 *  Adaptor to connect and execute queries on a Cassandra cluster. Uses pool
 *  connections.
 *
 *****************************************************************************/

/**
 * Module dependencies
 */
var adaptor = require('cassandra-client').PooledConnection;

/**
 * Cassandra
 *
 * @param {object] cluster
 * @param {string} keyspace
 * @constructor
 */
function Cassandra(cluster, keyspace) {

	/**
	 * Cassandra pool
	 * @type {*}
	 * @private
	 */
	this._pool = new adaptor({
		 hosts		: cluster.hosts
		,user		: cluster.user
		,pass		: cluster.pass
        ,maxSize	: this.DEFAULT_POOL_SIZE
		,keyspace	: keyspace
	});

	/**
	 * Connect to cassandra pool
	 * @type {*}
	 * @private
	 */
	this._pool.connect(function(error) {
		if (error) {
			this.kernel.emit(this.ERROR);
		} else {
			this.kernel.emit(this.CONNECTED);
		}
	}.bind(this));
}

/**
 * Connection events
 * @type {string}
 */
Cassandra.prototype.CONNECTED    = 'cassandra-connected';
Cassandra.prototype.DISCONNECTED = 'cassandra-disconnected';
Cassandra.prototype.ERROR        = 'cassandra-error';

/**
 * Status object
 */
Cassandra.prototype.status = {
	 querycnt : 0
	,running : {}
};

/**
 * Connection pool
 */
Cassandra.prototype._pool;
Cassandra.prototype.DEFAULT_POOL_SIZE = 10;

/**
 * Consistency constants
 * @type {string}
 */
Cassandra.prototype.CONSISTENCY_ONE			= 'ONE';
Cassandra.prototype.CONSISTENCY_ANY			= 'ANY';
Cassandra.prototype.CONSISTENCY_ALL			= 'ALL';
Cassandra.prototype.CONSISTENCY_QUORUM		= 'QUORUM';
Cassandra.prototype.CONSISTENCY_LOCAL_QUORUM= 'LOCAL_QUORUM';
Cassandra.prototype.CONSISTENCY_EACH_QUORUM	= 'EACH_QUORUM';

/**
 * Consistency properties for each query type
 * @type {string}
 */
Cassandra.prototype.CONSISTENCY_INSERT	= Cassandra.CONSISTENCY_ONE;
Cassandra.prototype.CONSISTENCY_SELECT	= Cassandra.CONSISTENCY_ONE;
Cassandra.prototype.CONSISTENCY_UPDATE	= Cassandra.CONSISTENCY_ONE;
Cassandra.prototype.CONSISTENCY_DELETE	= Cassandra.CONSISTENCY_ONE;
Cassandra.prototype.CONSISTENCY_BATCH	= Cassandra.CONSISTENCY_ONE;

/**
 * Executes a query on Cassandra Cluster
 *
 * @param {string} query
 * @param {array} params
 * @param {function} callback
 */
Cassandra.prototype.query = function(query, params, callback) {

	if (typeof query === 'undefined') {
		callback(new Error('Missing query'));
	} else if (typeof params === 'undefined') {
		callback(new Error('Missing binds'));
	} else {

		/**
		 * Add the query to the status object
		 * @type {*}
		 */
		var qid = Math.floor(Math.random() * 10000);

		this.status.querycnt++;
		this.status.running[qid] = {
			 query  : query
			,params : params
			,start  : this.kernel.utility.date('yyyy-mm-dd HH:ii:ss')
		};

		/**
		 * Set the consistency for this query
		 */
		query = this.setConsistency(query);

		/**
		 * Execute the query
		 */
		this._pool.execute(query, params, function(error, results) {
			if (error) {
				callback(error);
			} else {

				callback(null, results);

				/**
				 * Remove the query from the status object
				 */
				delete this.status.running[qid];
			}
		}.bind(this));
	}
};

/**
 * Add consistency for each query type
 *
 * @param {string} query
 * @return {string}
 */
Cassandra.prototype.setConsistency = function(query) {

	/**
	 * Check if consistency already set in the query
	 */
	if (query.toLowerCase().indexOf('consistency') != -1) {
		return query;
	}

	var hasWhere = query.toLowerCase().indexOf('where') != -1 ? true : false;
	var hasUsing = query.toLowerCase().indexOf('using') != -1 ? true : false;

	/**
	 * Add consistency to batch queries
	 */
	if (query.toLowerCase().indexOf('batch') != -1) {

		if (hasUsing) {
			return query.replace(
				/USING\s+TIMESTAMP\s+(.*?)\s+/i
				,'USING TIMESTAMP $1 AND CONSISTENCY ' + this.CONSISTENCY_BATCH + ' '
			);
		}

		return query.replace(/batch/i, 'BATCH USING CONSISTENCY ' + this.CONSISTENCY_BATCH);
	}

	/**
	 * Add consistency to select queries
	 */
	if (query.toLowerCase().indexOf('select') != -1) {

		if (hasWhere) {
			if (hasUsing) {
				return query.replace(/where/i, 'AND CONSISTENCY ' + this.CONSISTENCY_SELECT + ' WHERE');
			}

			return query.replace(/where/i, 'USING CONSISTENCY ' + this.CONSISTENCY_SELECT + ' WHERE');
		}

		if (hasUsing) {
			return query + ' AND CONSISTENCY ' + this.CONSISTENCY_SELECT;
		}

		return query + ' USING CONSISTENCY ' + this.CONSISTENCY_SELECT;
	}

	/**
	 * Add consistency to insert queries
	 */
	if (query.toLowerCase().indexOf('insert') != -1) {

		if (hasUsing) {
			return query + ' AND CONSISTENCY ' + this.CONSISTENCY_INSERT;
		}

		return query + ' USING CONSISTENCY ' + this.CONSISTENCY_INSERT;
	}

	/**
	 * Add consistency to update queries
	 */
	if (query.toLowerCase().indexOf('update') != -1) {

		if (hasUsing) {
			return query.replace(/set/i, 'AND CONSISTENCY ' + this.CONSISTENCY_UPDATE + ' SET');
		}

		return query.replace(/set/i, 'USING CONSISTENCY ' + this.CONSISTENCY_UPDATE + ' SET');
	}

	/**
	 * Add consistency to delete queries
	 */
	if (query.toLowerCase().indexOf('delete') != -1) {

		if (hasWhere) {
			if (hasUsing) {
				return query.replace(/where/i, 'AND CONSISTENCY ' + this.CONSISTENCY_DELETE + ' WHERE');
			}

			return query.replace(/where/i, 'USING CONSISTENCY ' + this.CONSISTENCY_DELETE + ' WHERE');
		}

		if (hasUsing) {
			return query + ' AND CONSISTENCY ' + this.CONSISTENCY_DELETE;
		}

		return query + ' USING CONSISTENCY ' + this.CONSISTENCY_DELETE;
	}

	/**
	 * Ignore consistency for other queries types(eg: ALTER, CREATE, DROP)
	 */
	return query;
};

/**
 * Shutting down the pool prevents further work from being enqueued,
 * and closes all open connections after pending requests are complete.
 * @param {function} cb
 */
Cassandra.prototype.shutdown = function(cb) {
	this._pool.shutdown(cb);
};

/**
 * Exports
 */
exports = module.exports = Cassandra;