/******************************************************************************
 * Lemonade JS
 * https://lemonadejs.com
 *
 * ----
 *
 * HTTP Daemon
 *
 *  Each HTTP request will create a new Instance object, attach http specific
 *  properties to it and do a URI based routing inside the controllers dir
 *
 *  Each controller will emit via Instance the usual action events (log, error,
 *  output, end) + two additional : flush and header
 *
 * @param {number} port
 *
 *****************************************************************************/

function Httpd (port) {
    this.port = port || 1337;
    this._server();
    this._listen();
}

/**
 * Module dependencies
 */
var  httplib  = require('http')
    ,url      = require('url')
    ,qs       = require('qs')
    ,util     = require('util');

/**
 * Event emitted when the httpd daemon is up
 * and listening for requests
 * @type {string}
 */
Httpd.prototype.ONLINE = 'online';

/**
 * Event emitted when the daemon is down
 * @type {string}
 */
Httpd.prototype.OFFLINE = 'offline';

/**
 * Httpd port
 * @type {number}
 */
Httpd.prototype.port;

/**
 * Status
 * @return {{running: {}, count: number, port: *}}
 */
Httpd.prototype.status = {
     running : {}
    ,count   : 0
    ,port    : this.port
};

/**
 * Internal server reference
 * @private
 */
Httpd.prototype._server;

/**
 * Idle connection timeout
 * @type {number}
 */
Httpd.prototype.timeout = 4000;

/**
 * Log file names (will be prefixed by kernel.log)
 * @private
 */
Httpd.prototype._accesslog = 'http.access';
Httpd.prototype._errorlog  = 'http.error';

/**
 * Overloaded instance prototype
 * @type {}
 * @private
 */
Httpd.prototype._instanceproto;

/**
 * Gracefully shutdown the server. All connections
 * will wait for httpd.timeout before closing
 * @return
 */
Httpd.prototype.shutdown = function(cb) {
    this._server.close(function(err) {
        this.kernel.emit(this.OFFLINE);
        cb(err);
    }.bind(this));
};

/**
 * Listen for various events
 * @private
 * @return
 */
Httpd.prototype._listen = function() {

    /**
     * Server is up and is receiving requests
     */
    this._server.addListener('listening', function() {
        this.kernel.emit(this.ONLINE);
    }.bind(this));

    /**
     *
     */
    this._server.addListener('connection',function(stream) {
        stream.setTimeout(this.timeout);
    }.bind(this));

};

/**
 * Creates the actual http server and defines the logic
 * behind request handling
 * @private
 * @return
 */
Httpd.prototype._server = function() {
    this._server = httplib.createServer(function (req, res) {

        /**
         * Check for post data
         * @type {string}
         */
        var post  = ''; req.on('data', function mergepost(chunk) {
            post += chunk.toString();
        });

        /**
         * Handle the request
         */
        req.on('end', function _handlerequest() {
            this._instance(req, res, post);
        }.bind(this));

    }.bind(this)).listen(this.port);
};

/**
 * Create a new Instance object for each request (overloaded with
 * http specific properties and methods)
 * @private
 * @param {*} req
 * @param {*} res
 * @param {string} post
 * @return
 */
Httpd.prototype._instance = function (req, res, post) {

    /**
     * Overload the instance with our instance reloaded
     */
    if (!this._instanceproto) {
        this._instanceproto = this.kernel.include(this.kernel.kerneldir + '/instance')
                            . extends(this.instancereloaded);
    }

    var instanceobj = this._instanceproto.new();
    instanceobj.id  = 'http-' + this.status.count;

    /**
     * Bind http request and response
     */
    instanceobj.request  = req;
    instanceobj.response = res;

    /**
     * get / post shortcuts
     */
    instanceobj.get  = url.parse(instanceobj.request.url, true).query;
    instanceobj.post = {};

    if (post) {
        try {
            instanceobj.post = JSON.parse(post);
        } catch (e) {
            instanceobj.post = qs.parse(post);
        }
    }

    /**
     * Internals
     */
    instanceobj._outputbuffer = '';
    instanceobj._httpserver   = this;
    instanceobj._header       = { code: 200, entries: {'Content-type': 'text/json'} };
    instanceobj._headersent   = false;
    instanceobj._flushed      = false;
    
    /**
     * Override the instance id to be http specific
     */
    this.status.count++;

    /**
     * Status info
     */
    this.status.running[instanceobj.id] = {

         id     : instanceobj.id
        ,req    : instanceobj.request.url
        ,start  : this.kernel.utility.date('yyyy-mm-dd HH:ii:ss', instanceobj.starttime)
        ,method : instanceobj.request.method
        ,ip     : instanceobj.request.connection.remoteAddress

    };

    /**
     * Route the request to the appropriate controller
     */
    instanceobj.route();
};

/**
 * Placeholder
 */
Httpd.prototype.instancereloaded = function() {};

/**
 * Route the request to the appropriate controller based
 * on URI. Only first dir applies, then that controller must
 * handle routing from there.
 * @return
 */
Httpd.prototype.instancereloaded.route = function() {

    var  arr  = url.parse(this.request.url).pathname.split('/')
        ,def  = 'default'
        ,path = def;

    if (arr[1]) {
        path = arr[1];
    }

    /**
     * Try to load the module, if not found
     * will throw
     */
    try {
        this.controller(path).new();
    } catch (err) {
        /**
         * Module not found, this is 404
         */
        if (err.code === 'MODULE_NOT_FOUND') {
            this.end(
                new Error('Invalid route')
            );
        } else {
            this.end(err);
        }
    }
};

/**
 * Mark the instance as ended, flush if needed and remove from
 * the internal scoreboard. Log to the appropriate file depending
 * on error
 * @return
 */
Httpd.prototype.instancereloaded.end = function(error, data) {

    if (!this._flushed) {
        this.flush(error, data);
    }

    delete this._httpserver.status.running[this.id];

    var  duration = ((this.kernel.utility.microtime()
                    - this.kernel.utility.microtime(this.starttime)) / 1000) || 0

        ,logmsg   = (this.request.headers['x-forwarded-for'] || this.request.connection.remoteAddress) + ' '
                  +  this.request.method + ' '
                  +  this.request.url + ' '
                  +  duration.toFixed(3) + 's';

    if (error) {
        this.kernel.log(logmsg + ' ' + error, this._httpserver._errorlog, this.starttime);
    } else {
        this.kernel.log(logmsg, this._httpserver._accesslog, this.starttime);
    }
};

/**
 * Flush the buffer (or data) via http.response and end the
 * response. Send the headers if not sent.
 *
 * @param {string|Error} error
 * @param {*} data
 * @return
 */
Httpd.prototype.instancereloaded.flush = function(error, data) {

    if (error instanceof Error) {
        error = error.message;
    }

    if (data && this._outputbuffer) {
        this._outputbuffer += data;
    } else if (data) {
        this._outputbuffer  = data;
    }

    var duration = ((this.kernel.utility.microtime() - this.starttime) / 1000) || 0;

    if (!this._headersent) {
        this.header();
    }

    if (this._header.entries['Content-type'] === 'text/json') {
        this.response.write(
            JSON.stringify({
                'error'    : !!error
                ,'duration' : duration
                ,'response' : (error) ? error : this._outputbuffer
            })
        );
    } else {
        this.response.write(
            (error || this._outputbuffer) + ''
        );
    }

    this.response.end();
    this._flushed = true;
};

/**
 * Sends the http headers and response code
 *
 * @param {number} code
 * @param {object} entries
 * @return
 */
Httpd.prototype.instancereloaded.header = function(code, entries) {

    if (entries) {
        this._header.entries = entries;
    }

    if (code) {
        this._header.code = code;
    }

    this._headersent = true;
    this.response.writeHead(this._header.code, this._header.entries);
};

/**
 * Append data to the outputbuffer
 *
 * @param {string} output
 * @return
 */
Httpd.prototype.instancereloaded.echo = function(output) {
    this._outputbuffer += output;
};

/**
 * Log a message to the http error log
 *
 * @param {string|Error} object
 * @return
 */
Httpd.prototype.instancereloaded.error = function(object) {
    this.kernel.log(object, this._errorlog);
};

/**
 * Log a message to the http log
 *
 * @param {string} message
 * @return
 */
Httpd.prototype.instancereloaded.log = function(message) {
    this.kernel.log(message, this._accesslog);
};

/**
 * Exports
 */
exports = module.exports = Httpd;